package ru.dbotthepony.mc.otm.block

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.particles.DustParticleOptions
import net.minecraft.util.RandomSource
import net.minecraft.world.Container
import net.minecraft.world.Containers
import net.minecraft.world.InteractionHand
import net.minecraft.world.InteractionResult
import net.minecraft.world.MenuProvider
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.context.BlockPlaceContext
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.EntityBlock
import net.minecraft.world.level.block.Rotation
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.state.StateDefinition
import net.minecraft.world.level.block.state.properties.EnumProperty
import net.minecraft.world.level.material.Material
import net.minecraft.world.level.material.MaterialColor
import net.minecraft.world.phys.BlockHitResult
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.WorkerState
import ru.dbotthepony.mc.otm.core.component1
import ru.dbotthepony.mc.otm.core.component2
import ru.dbotthepony.mc.otm.core.component3
import ru.dbotthepony.mc.otm.core.get

interface IDroppableContainer {
	val droppableContainer: Container

	fun beforeDroppingItems(
		oldBlockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		newBlockState: BlockState,
		movedByPiston: Boolean
	) {}

	fun beforeDestroyedByPlayer(level: Level, blockPos: BlockPos, blockState: BlockState, player: Player) {}
}

abstract class MatteryBlock @JvmOverloads constructor(
	properties: Properties = DEFAULT_PROPERTIES
) : Block(properties) {
	override fun setPlacedBy(
		level: Level,
		blockPos: BlockPos,
		blockState: BlockState,
		entity: LivingEntity?,
		itemStack: ItemStack
	) {
		if (this is EntityBlock && itemStack.hasCustomHoverName() && !level.isClientSide) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is MatteryBlockEntity)
				tile.customDisplayName = itemStack.displayName
		}

		super.setPlacedBy(level, blockPos, blockState, entity, itemStack)
	}

	@Suppress("OVERRIDE_DEPRECATION")
	override fun use(
		blockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		ply: Player,
		hand: InteractionHand,
		blockHitResult: BlockHitResult
	): InteractionResult {
		if (this is EntityBlock && !level.isClientSide) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is MenuProvider) {
				ply.openMenu(tile)
				return InteractionResult.CONSUME
			}
		}

		if (this is EntityBlock && level.isClientSide)
			return InteractionResult.SUCCESS

		return super.use(blockState, level, blockPos, ply, hand, blockHitResult)
	}

	override fun animateTick(blockState: BlockState, level: Level, blockPos: BlockPos, random: RandomSource) {
		if (this is EntityBlock) {
			val state = blockState.getOptionalValue(WorkerState.WORKER_STATE).or { blockState.getOptionalValue(WorkerState.SEMI_WORKER_STATE) }

			if (state.isPresent && state.get() == WorkerState.WORKING) {
				val state2 = blockState.getOptionalValue(RotatableMatteryBlock.FACING).or { blockState.getOptionalValue(RotatableMatteryBlock.FACING_FULL) }

				if (state2.isPresent) {
					val direction = state2.get()

					val (x, y, z) = blockPos
					var xd = x + 0.5
					var yd = y + 0.5
					var zd = z + 0.5

					val (nx, ny, nz) = direction.normal
					xd += nx * 0.5
					yd += ny * 0.5
					zd += nz * 0.5

					when (direction) {
						Direction.DOWN, Direction.UP -> {
							xd += random.nextDouble() - 0.5
							zd += random.nextDouble() - 0.5
						}

						Direction.NORTH, Direction.SOUTH -> {
							xd += random.nextDouble() - 0.5
							yd += random.nextDouble() - 0.5
						}

						Direction.WEST, Direction.EAST -> {
							yd += random.nextDouble() - 0.5
							zd += random.nextDouble() - 0.5
						}
					}

					level.addParticle(DustParticleOptions.REDSTONE, xd, yd, zd, 0.0, 0.0, 0.0)
				}
			}
		}
	}

	@Suppress("OVERRIDE_DEPRECATION")
	override fun getMenuProvider(blockState: BlockState, level: Level, blockPos: BlockPos): MenuProvider? {
		if (this is EntityBlock) {
			val tile = level.getBlockEntity(blockPos)

			if (tile is MenuProvider) {
				return tile
			}
		}

		return null
	}

	@Suppress("OVERRIDE_DEPRECATION")
	override fun neighborChanged(
		state: BlockState,
		level: Level,
		pos: BlockPos,
		neighbour: Block,
		neighbourPos: BlockPos,
		movedByPiston: Boolean
	) {
		super.neighborChanged(state, level, pos, neighbour, neighbourPos, movedByPiston)

		if (this is EntityBlock && !level.isClientSide) {
			val tile = level.getBlockEntity(pos)

			if (tile is MatteryBlockEntity)
				tile.redstoneSignal = level.getBestNeighborSignal(pos)
		}
	}

	@Suppress("OVERRIDE_DEPRECATION")
	override fun onRemove(
		oldBlockState: BlockState,
		level: Level,
		blockPos: BlockPos,
		newBlockState: BlockState,
		movedByPiston: Boolean
	) {
		if (!oldBlockState.`is`(newBlockState.block)) {
			val blockentity = level.getBlockEntity(blockPos)

			if (blockentity is IDroppableContainer) {
				blockentity.beforeDroppingItems(oldBlockState, level, blockPos, newBlockState, movedByPiston)
				Containers.dropContents(level, blockPos, blockentity.droppableContainer)
				level.updateNeighbourForOutputSignal(blockPos, this)
			}
		}

		@Suppress("DEPRECATION")
		super.onRemove(oldBlockState, level, blockPos, newBlockState, movedByPiston)
	}

	override fun playerWillDestroy(level: Level, blockPos: BlockPos, blockState: BlockState, player: Player) {
		super.playerWillDestroy(level, blockPos, blockState, player)

		val blockentity = level.getBlockEntity(blockPos)

		if (blockentity is IDroppableContainer) {
			blockentity.beforeDestroyedByPlayer(level, blockPos, blockState, player)
		}
	}

	companion object {
		val DEFAULT_PROPERTIES: Properties = Properties.of(Material.STONE, MaterialColor.METAL).requiresCorrectToolForDrops().strength(1.5f, 25.0f)
	}
}

abstract class RotatableMatteryBlock @JvmOverloads constructor(properties: Properties = DEFAULT_PROPERTIES) : MatteryBlock(properties) {
	init {
		@Suppress("LeakingThis")
		registerDefaultState(getStateDefinition().any().setValue(facingProperty, Direction.SOUTH))
	}

	override fun createBlockStateDefinition(builder: StateDefinition.Builder<Block, BlockState>) {
		builder.add(facingProperty)
	}

	override fun getStateForPlacement(context: BlockPlaceContext): BlockState? {
		if (hasFreeRotation) {
			return defaultBlockState().setValue(
				FACING_FULL,
				if (faceToPlayer(context)) context.nearestLookingDirection.opposite else context.nearestLookingDirection
			)
		} else {
			return defaultBlockState().setValue(
				FACING,
				if (faceToPlayer(context)) context.horizontalDirection.opposite else context.horizontalDirection
			)
		}
	}

	@Suppress("OVERRIDE_DEPRECATION")
	override fun rotate(blockState: BlockState, rotation: Rotation): BlockState {
		@Suppress("DEPRECATION")
		return super.rotate(blockState, rotation).setValue(facingProperty, rotation.rotate(blockState[facingProperty]))
	}

	val facingProperty get() = if (hasFreeRotation) FACING_FULL else FACING
	open val hasFreeRotation get() = false

	open fun faceToPlayer(context: BlockPlaceContext): Boolean {
		return true
	}

	companion object {
		val FACING: EnumProperty<Direction> = EnumProperty.create(
			"facing",
			Direction::class.java,
			Direction.SOUTH,
			Direction.WEST,
			Direction.NORTH,
			Direction.EAST
		)

		val FACING_FULL: EnumProperty<Direction> = EnumProperty.create("facing", Direction::class.java)
	}
}
