package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.ChatFormatting
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.Block
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.map

private fun isReason(status: Any?, reason: Any) = status == null || status == reason

abstract class MatteryWorkerBlockEntity<JobType : MatteryWorkerBlockEntity.Job>(
	type: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	val jobDeserializer: (tag: CompoundTag) -> JobType?
) : MatteryPoweredBlockEntity(type, blockPos, blockState) {
	open class Job {
		open val ticks: Double
		open val powerUsage: Decimal

		constructor(
			ticks: Double,
			powerUsage: Decimal = Decimal.ZERO
		) {
			this.ticks = ticks
			this.powerUsage = powerUsage
		}

		constructor(
			tag: CompoundTag
		) : this(tag.getDouble(TICKS_KEY), tag.getDecimal(POWER_KEY))

		open fun serializeNBT(): CompoundTag {
			return CompoundTag().also {
				it[TICKS_KEY] = ticks
				it[POWER_KEY] = powerUsage
			}
		}

		companion object {
			const val TICKS_KEY = "ticks"
			const val POWER_KEY = "power"
		}
	}

	@Suppress("LeakingThis")
	open class ItemJob : Job {
		open val itemStack: ItemStack

		constructor(
			itemStack: ItemStack,
			ticks: Double,
			power: Decimal = Decimal.ZERO
		) : super(ticks, power) {
			this.itemStack = itemStack
		}

		constructor(
			tag: CompoundTag
		) : super(tag) {
			this.itemStack = (tag["item"] as? CompoundTag)?.let { ItemStack.of(it) } ?: ItemStack.EMPTY
		}

		override fun serializeNBT(): CompoundTag {
			return super.serializeNBT().also {
				it[ITEM_KEY] = itemStack.serializeNBT()
			}
		}

		companion object {
			const val ITEM_KEY = "item"
		}
	}

	enum class IdleReason {
		ITEM,
		POWER,
		MATTER,

		/**
		 * Observing external factor, such as waiting for matter/item network tasks
		 */
		OBSERVING
	}

	data class Status(
		val success: Boolean,
		val throttleTicks: Int = 0,
		val idleReason: IdleReason? = null,
		val newDrainedPower: Decimal? = null
	) {
		companion object {
			val SUCCESS = Status(true)
			val FAILURE = Status(false)
			val FAILURE_ITEM = Status(false, 20, IdleReason.ITEM)
			val FAILURE_MATTER = Status(false, 20, IdleReason.MATTER)
			val FAILURE_WAIT = Status(false, 100)
			val FAILURE_WAIT_FAST = Status(false, 20)
		}
	}

	var workTicks = 0.0
		protected set

	var throttleTicks = 0
		protected set

	protected open fun jobUpdated(oldJob: JobType?, newJob: JobType?) {}

	var currentJob: JobType? = null
		protected set(value) {
			if (field != value) {
				val old = field
				field = value
				jobUpdated(old, value)
			}
		}

	/**
	 * Can be whatever you want, but [IdleReason] certainly contains all cases
	 */
	var idleReason: Any? = null
		private set

	var isIdling = false
		protected set

	val isUnableToProcess: Boolean get() = throttleTicks > 0

	val workProgress: Float
		get() {
			val currentJob = currentJob ?: return 0.0f
			return (workTicks / currentJob.ticks).coerceAtMost(1.0).toFloat()
		}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt[WORK_TICKS_KEY] = workTicks
		currentJob?.let { nbt[JOB_KEY] = it.serializeNBT() }
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		workTicks = nbt.getDouble(WORK_TICKS_KEY)
		currentJob = nbt.map(JOB_KEY, jobDeserializer::invoke)

		if (currentJob == null)
			workTicks = 0.0
	}

	override fun setChanged() {
		super.setChanged()
		isIdling = false
	}

	override fun setChangedLight() {
		super.setChangedLight()
		isIdling = false
	}

	protected fun powerLevelUpdated() {
		super.setChangedLight()

		if (isReason(idleReason, IdleReason.POWER)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	protected fun itemContainerUpdated() {
		super.setChanged()

		if (isReason(idleReason, IdleReason.ITEM)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	protected fun matterLevelUpdated() {
		super.setChangedLight()

		if (isReason(idleReason, IdleReason.MATTER)) {
			isIdling = false
			throttleTicks = 0
		}
	}

	/**
	 * Called whenever reaching desired amount of ticks at job
	 */
	protected abstract fun onJobFinish(job: JobType): Status

	/**
	 * [Pair.second] is reason why job can't be performed
	 *
	 * If not null, it is written to [idleReason]
	 */
	protected abstract fun computeNextJob(): Pair<JobType?, Any?>

	protected open fun onWorkTick(requiredPower: Decimal, extractedPower: Decimal, ticksAdvanced: Double, job: JobType): Status {
		return Status.SUCCESS
	}

	private var idleTicksAnim = 0
	private var workingTicksAnim = 0
	private var errorTicksAnim = 0

	override fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {
		super.redstoneStatusUpdated(newBlocked, oldBlocked)
		isIdling = newBlocked
	}

	protected fun workerLoop() {
		if (errorTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.ERROR) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS)
		}

		if (throttleTicks > 0) {
			workingTicksAnim = 0
			idleTicksAnim = 0
			throttleTicks--
			errorTicksAnim++

			if (throttleTicks > 0)
				return
			else
				isIdling = false
		}

		if (isIdling) {
			workingTicksAnim = 0
			errorTicksAnim = 0
			idleTicksAnim++

			if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(
					WorkerState.WORKER_STATE
				) != WorkerState.IDLE
			) {
				level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
			}

			return
		}

		var availableTicks = 1.0

		while (!isIdling && weakGreaterThan(availableTicks, 0.0) && throttleTicks <= 0) {
			if (isBlockedByRedstone) {
				isIdling = true
				break
			}

			var currentJob = currentJob

			if (currentJob == null) {
				val (job, reason) = computeNextJob()

				if (job == null) {
					idleReason = reason
					isIdling = reason != null
					workingTicksAnim = 0
					break
				}

				this.currentJob = job
				currentJob = job
			}

			if (!currentJob.powerUsage.isZero && energy.batteryLevel.isZero) {
				idleReason = IdleReason.POWER
				isIdling = true
				idleTicksAnim++

				if (idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(
						WorkerState.WORKER_STATE
					) != WorkerState.IDLE
				) {
					level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
				}

				break
			}

			idleTicksAnim = 0

			if (weakLessThan(workTicks, currentJob.ticks)) {
				val ticksLeft = currentJob.ticks - workTicks
				val ticksAdvanced: Double

				var requiredPower: Decimal? = null
				var extractedPower: Decimal? = null

				if (currentJob.powerUsage.isZero) {
					ticksAdvanced = availableTicks.coerceAtMost(ticksLeft)
				} else {
					requiredPower = currentJob.powerUsage * ticksLeft.coerceAtMost(availableTicks)
					extractedPower = energy.extractEnergyInner(requiredPower, true)
					ticksAdvanced = (extractedPower / requiredPower).toDouble().coerceAtMost(ticksLeft).coerceAtMost(availableTicks)
				}

				if (weakEqualDoubles(ticksAdvanced, 0.0)) {
					break
				}

				val status = onWorkTick(requiredPower ?: Decimal.ZERO, extractedPower ?: Decimal.ZERO, ticksAdvanced, currentJob)

				if (!status.success) {
					throttleTicks += status.throttleTicks

					if (status.idleReason != null) {
						idleReason = status.idleReason
						isIdling = true
					}

					break
				}

				workingTicksAnim++
				errorTicksAnim = 0

				workTicks += ticksAdvanced
				availableTicks -= ticksAdvanced

				extractedPower = status.newDrainedPower ?: extractedPower

				if (extractedPower != null) {
					energy.extractEnergyInner(extractedPower, false)
				}

				continue
			}

			val status = onJobFinish(currentJob)

			if (status.success) {
				this.currentJob = null
				workTicks = 0.0
				errorTicksAnim = 0
			} else {
				throttleTicks += status.throttleTicks

				if (status.idleReason != null) {
					idleReason = status.idleReason
					isIdling = true
				}

				errorTicksAnim++
				break
			}
		}

		if (workingTicksAnim > 20 &&
			errorTicksAnim == 0 &&
			blockState.hasProperty(WorkerState.WORKER_STATE) &&
			blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.WORKING
		) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
		}
	}

	fun basicTicker() {
		batteryChargeLoop()
		workerLoop()
	}

	companion object {
		const val WORK_TICKS_KEY = "workTicks"
		const val JOB_KEY = "job"

		fun appendHoverText(itemStack: ItemStack, blockGetter: BlockGetter?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			val tag = itemStack.tag ?: return

			val subtag = tag.get("BlockEntityTag") as? CompoundTag

			if (subtag != null) {
				if (subtag.contains(WORK_TICKS_KEY) && !subtag.contains(JOB_KEY)) {
					if (subtag.getDouble(WORK_TICKS_KEY) != 0.0)
						tooltips.add(TranslatableComponent("otm.item.worker.work_ticks_mono", "%.1f".format(subtag.getDouble(WORK_TICKS_KEY))).withStyle(ChatFormatting.GRAY))
				} else if (subtag.contains(WORK_TICKS_KEY) && tag.contains(JOB_KEY)) {
					tooltips.add(TranslatableComponent("otm.item.worker.work_ticks",
						"%.1f".format(subtag.getDouble(WORK_TICKS_KEY)),
						"%.1f".format(Job(subtag.getCompound(JOB_KEY)).ticks)).withStyle(ChatFormatting.GRAY))
				}
			}

			WorkerEnergyStorage.appendHoverText(itemStack, blockGetter, tooltips, flag)
			MatteryPoweredBlockEntity.appendHoverText(itemStack, blockGetter, tooltips, flag)
		}
	}
}
