package ru.dbotthepony.mc.otm.core

import net.minecraft.world.item.ItemStack
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.io.OutputStream
import java.util.*
import java.util.function.Predicate
import kotlin.NoSuchElementException
import kotlin.math.absoluteValue

/**
 * Represents value which can be encoded onto or decoded from stream.
 *
 * Also provides [copy] and [compare] methods
 */
interface IStreamCodec<V> {
	fun read(stream: DataInputStream): V
	fun write(stream: DataOutputStream, value: V)

	/**
	 * if value is immutable, return it right away
	 */
	fun copy(value: V): V

	/**
	 * Optional equality check override. Utilized to determine whenever e.g. network value is different from new value
	 *
	 * By default uses [Any.equals]
	 */
	fun compare(a: V, b: V): Boolean {
		return a == b
	}
}

class StreamCodec<V>(
	private val reader: (stream: DataInputStream) -> V,
	private val writer: (stream: DataOutputStream, value: V) -> Unit,
	private val copier: ((value: V) -> V) = { it },
	private val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
) : IStreamCodec<V> {
	override fun read(stream: DataInputStream): V {
		return reader.invoke(stream)
	}

	override fun write(stream: DataOutputStream, value: V) {
		writer.invoke(stream, value)
	}

	override fun copy(value: V): V {
		return copier.invoke(value)
	}

	override fun compare(a: V, b: V): Boolean {
		return comparator.invoke(a, b)
	}
}

val NullValueCodec = StreamCodec({ null }, { _, _ -> })
val BooleanValueCodec = StreamCodec(DataInputStream::readBoolean, DataOutputStream::writeBoolean)
val ByteValueCodec = StreamCodec(DataInputStream::readByte, { s, v -> s.writeByte(v.toInt()) })
val ShortValueCodec = StreamCodec(DataInputStream::readShort, { s, v -> s.writeShort(v.toInt()) })
val IntValueCodec = StreamCodec(DataInputStream::readInt, DataOutputStream::writeInt)
val LongValueCodec = StreamCodec(DataInputStream::readLong, DataOutputStream::writeLong)
val FloatValueCodec = StreamCodec(DataInputStream::readFloat, DataOutputStream::writeFloat)
val DoubleValueCodec = StreamCodec(DataInputStream::readDouble, DataOutputStream::writeDouble)
val ItemStackValueCodec = StreamCodec(DataInputStream::readItem, DataOutputStream::writeItem, ItemStack::copy) { a, b -> a.equals(b, true) }
val ImpreciseFractionValueCodec = StreamCodec(DataInputStream::readDecimal, DataOutputStream::writeDecimal)
val BigDecimalValueCodec = StreamCodec(DataInputStream::readBigDecimal, DataOutputStream::writeBigDecimal)
val UUIDValueCodec = StreamCodec({ s -> UUID(s.readLong(), s.readLong()) }, { s, v -> s.writeLong(v.mostSignificantBits); s.writeLong(v.leastSignificantBits) })
val VarIntValueCodec = StreamCodec(DataInputStream::readVarIntLE, DataOutputStream::writeVarIntLE)
val VarLongValueCodec = StreamCodec(DataInputStream::readVarLongLE, DataOutputStream::writeVarLongLE)

class EnumValueCodec<V : Enum<V>>(private val clazz: Class<out V>) : IStreamCodec<V> {
	private val values = clazz.enumConstants

	override fun read(stream: DataInputStream): V {
		val id = stream.readVarIntLE()

		if (id >= values.size) {
			throw IndexOutOfBoundsException("Unable to read enum $clazz, no such enum at index $id")
		}

		return values[id]
	}

	override fun write(stream: DataOutputStream, value: V) {
		stream.writeVarIntLE(value.ordinal)
	}

	override fun copy(value: V): V {
		return value
	}

	override fun compare(a: V, b: V): Boolean {
		return a === b
	}
}

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutputStream) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun InputStream.readInt(): Int {
	if (this is DataInputStream) {
		return readInt()
	}

	return (read() shl 24) or (read() shl 16) or (read() shl 8) or read()
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutputStream) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun InputStream.readLong(): Long {
	if (this is DataInputStream) {
		return readLong()
	}

	return (read().toLong() shl 48) or
		(read().toLong() shl 40) or
		(read().toLong() shl 32) or
		(read().toLong() shl 24) or
		(read().toLong() shl 16) or
		(read().toLong() shl 8) or
		read().toLong()
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun InputStream.readFloat() = Float.fromBits(readInt())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())
fun InputStream.readDouble() = Double.fromBits(readLong())

fun InputStream.readVarIntLE(): Int {
	val readFirst = read()

	if (readFirst < 0) {
		throw NoSuchElementException("Reached end of stream")
	}

	if (readFirst and 64 == 0) {
		return if (readFirst and 128 != 0) -(readFirst and 63) else readFirst and 63
	}

	var result = 0
	var nextBit = readFirst and 64
	var read = readFirst and 63
	var i = 0

	while (nextBit != 0) {
		result = result or (read shl i)
		read = read()

		if (read < 0) {
			throw NoSuchElementException("Reached end of stream")
		}

		nextBit = read and 128
		read = read and 127

		if (i == 0)
			i = 6
		else
			i += 7
	}

	result = result or (read shl i)
	return if (readFirst and 128 != 0) -result else result
}

fun OutputStream.writeVarIntLE(value: Int) {
	write((if (value < 0) 128 else 0) or (if (value in -63 .. 63) 0 else 64) or (value.absoluteValue and 63))
	var written = value.absoluteValue ushr 6

	while (written != 0) {
		write((written and 127) or (if (written >= 128) 128 else 0))
		written = written ushr 7
	}
}

fun InputStream.readVarLongLE(): Long {
	val readFirst = read()

	if (readFirst < 0) {
		throw NoSuchElementException("Reached end of stream")
	}

	if (readFirst and 64 == 0) {
		return if (readFirst and 128 != 0) -(readFirst and 63).toLong() else (readFirst and 63).toLong()
	}

	var result = 0L
	var nextBit = readFirst and 64
	var read = readFirst and 63
	var i = 0

	while (nextBit != 0) {
		result = result or (read shl i).toLong()
		read = read()

		if (read < 0) {
			throw NoSuchElementException("Reached end of stream")
		}

		nextBit = read and 128
		read = read and 127

		if (i == 0)
			i = 6
		else
			i += 7
	}

	result = result or (read shl i).toLong()
	return if (readFirst and 128 != 0) -result else result
}

fun OutputStream.writeVarLongLE(value: Long) {
	write((if (value < 0L) 128 else 0) or (if (value in -63 .. 63) 0 else 64) or (value.absoluteValue and 63).toInt())
	var written = value.absoluteValue ushr 6

	while (written != 0L) {
		write((written and 127).toInt() or (if (written >= 128) 128 else 0))
		written = written ushr 7
	}
}

private data class IndexedStreamCodec<T>(
	val condition: Predicate<Any?>,
	val id: Int,
	val codec: StreamCodec<T>
) {
	fun read(stream: DataInputStream): T {
		return codec.read(stream)
	}

	fun write(stream: DataOutputStream, value: Any?) {
		codec.write(stream, value as T)
	}
}

private inline fun <reified T> IndexedStreamCodec(id: Int, codec: StreamCodec<T>): IndexedStreamCodec<T> {
	return IndexedStreamCodec({ it is T }, id, codec)
}

private val codecs: List<IndexedStreamCodec<*>> = com.google.common.collect.ImmutableList.builder<IndexedStreamCodec<*>>().let {
	var codecID = 0

	it.add(IndexedStreamCodec({ it == null }, codecID++, NullValueCodec))

	it.add(IndexedStreamCodec(codecID++, BooleanValueCodec))
	it.add(IndexedStreamCodec(codecID++, ByteValueCodec))
	it.add(IndexedStreamCodec(codecID++, ShortValueCodec))
	it.add(IndexedStreamCodec(codecID++, VarIntValueCodec))
	it.add(IndexedStreamCodec(codecID++, VarLongValueCodec))
	it.add(IndexedStreamCodec(codecID++, FloatValueCodec))
	it.add(IndexedStreamCodec(codecID++, DoubleValueCodec))
	it.add(IndexedStreamCodec(codecID++, ItemStackValueCodec))
	it.add(IndexedStreamCodec(codecID++, ImpreciseFractionValueCodec))
	it.add(IndexedStreamCodec(codecID++, BigDecimalValueCodec))
	it.add(IndexedStreamCodec(codecID++, UUIDValueCodec))

	it.build()
}

class NotSerializableValueException(message: String) : Exception(message)

/**
 * Write arbitrary data to this stream, in exploit-free way
 */
fun DataOutputStream.writeType(value: Any?) {
	for (codec in codecs) {
		if (codec.condition.test(value)) {
			write(codec.id)
			codec.write(this, value)
			return
		}
	}

	throw NotSerializableValueException("Value $value <${value?.let { it::class.qualifiedName }}> can not be serialized")
}

/**
 * Read arbitrary data from this stream, in exploit-free way
 */
fun DataInputStream.readType(): Any? {
	val id = read()

	if (id >= codecs.size) {
		throw IndexOutOfBoundsException("No codec for network type $id")
	}

	return codecs[id].read(this)
}
