package ru.dbotthepony.mc.otm.data.condition

import com.google.gson.JsonDeserializationContext
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSerializationContext
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.Serializer
import net.minecraft.world.level.storage.loot.parameters.LootContextParams
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraft.world.level.storage.loot.predicates.LootItemConditionType
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.data.get
import ru.dbotthepony.mc.otm.registry.MLootItemConditions

data class ChanceWithPlaytimeCondition(
	val minPlaytime: Int = 0,
	val maxPlaytime: Int,
	val minProbability: Double,
	val maxProbability: Double
) : LootItemCondition, LootItemCondition.Builder {
	init {
		require(minPlaytime <= maxPlaytime) { "Invalid playtime range: min $minPlaytime, max $maxPlaytime" }
		require(minProbability <= maxProbability) { "Invalid probability range: min $minProbability, max $maxProbability" }

		require(maxProbability <= 1.0) { "Excessive max probability: $maxProbability" }
	}

	override fun test(t: LootContext): Boolean {
		return t[LootContextParams.LAST_DAMAGE_PLAYER]?.matteryPlayer?.let {
			val chance: Double

			if (it.ticksIExist < minPlaytime) {
				chance = minProbability
			} else if (it.ticksIExist >= maxPlaytime) {
				chance = maxProbability
			} else {
				chance = minProbability + (maxProbability - minProbability) * (it.ticksIExist - minPlaytime).toDouble() / (maxPlaytime - minPlaytime).toDouble()
			}

			return t.random.nextDouble() <= chance
		} ?: (t.random.nextDouble() <= minProbability)
	}

	override fun getType(): LootItemConditionType {
		return MLootItemConditions.CHANCE_WITH_PLAYTIME
	}

	override fun build(): LootItemCondition {
		return this
	}

	companion object : Serializer<ChanceWithPlaytimeCondition> {
		override fun serialize(
			p_79325_: JsonObject,
			p_79326_: ChanceWithPlaytimeCondition,
			p_79327_: JsonSerializationContext
		) {
			p_79325_["minPlaytime"] = JsonPrimitive(p_79326_.minPlaytime)
			p_79325_["maxPlaytime"] = JsonPrimitive(p_79326_.maxPlaytime)
			p_79325_["minProbability"] = JsonPrimitive(p_79326_.minProbability)
			p_79325_["maxProbability"] = JsonPrimitive(p_79326_.maxProbability)
		}

		override fun deserialize(
			p_79323_: JsonObject,
			p_79324_: JsonDeserializationContext
		): ChanceWithPlaytimeCondition {
			return ChanceWithPlaytimeCondition(
				minPlaytime = p_79323_["minPlaytime"].asInt,
				maxPlaytime = p_79323_["maxPlaytime"].asInt,
				minProbability = p_79323_["minProbability"].asDouble,
				maxProbability = p_79323_["maxProbability"].asDouble,
			)
		}
	}
}
