package ru.dbotthepony.mc.otm.matter

import com.google.gson.JsonParseException
import ru.dbotthepony.mc.otm.core.Decimal
import ru.dbotthepony.mc.otm.core.integerDivisionDown
import ru.dbotthepony.mc.otm.core.integerDivisionUp

enum class MatterFunction {
	ADD {
		override fun updateValue(self: Int, other: Int): Int = 												self + other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self + other
		override fun updateValue(self: Double, other: Double): Double = 									self + other
	},
	SUBTRACT {
		override fun updateValue(self: Int, other: Int): Int = 												self - other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self - other
		override fun updateValue(self: Double, other: Double): Double = 									self - other
	},
	MULTIPLY {
		override fun updateValue(self: Int, other: Int): Int = 												self * other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self * other
		override fun updateValue(self: Double, other: Double): Double = 									self * other
	},
	DIVIDE {
		override fun updateValue(self: Int, other: Int): Int = 												self / other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self / other
		override fun updateValue(self: Double, other: Double): Double = 									self / other
	},
	DIVIDE_UP {
		override fun updateValue(self: Int, other: Int): Int = integerDivisionUp(self, other)
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	throw JsonParseException("Integer division up is available only for integers")
		override fun updateValue(self: Double, other: Double): Double = 									throw JsonParseException("Integer division up is available only for integers")
	},
	DIVIDE_DOWN {
		override fun updateValue(self: Int, other: Int): Int = integerDivisionDown(self, other)
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	throw JsonParseException("Integer division down is available only for integers")
		override fun updateValue(self: Double, other: Double): Double = 									throw JsonParseException("Integer division down is available only for integers")
	},
	MODULO {
		override fun updateValue(self: Int, other: Int): Int = 												self % other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self % other
		override fun updateValue(self: Double, other: Double): Double = 									self % other
	},
	AT_LEAST {
		override fun updateValue(self: Int, other: Int): Int = 												self.coerceAtLeast(other)
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self.coerceAtLeast(other)
		override fun updateValue(self: Double, other: Double): Double = 									self.coerceAtLeast(other)
	},
	AT_MOST {
		override fun updateValue(self: Int, other: Int): Int = 												self.coerceAtMost(other)
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self.coerceAtMost(other)
		override fun updateValue(self: Double, other: Double): Double = 									self.coerceAtMost(other)
	},
	REPLACE {
		override fun updateValue(self: Int, other: Int): Int = 												other
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	other
		override fun updateValue(self: Double, other: Double): Double = 									other
	},
	NOOP {
		override fun updateValue(self: Int, other: Int): Int = 												self
		override fun updateValue(self: Decimal, other: Decimal): Decimal = 	self
		override fun updateValue(self: Double, other: Double): Double = 									self
	},
	;

	protected abstract fun updateValue(self: Int, other: Int): Int
	protected abstract fun updateValue(self: Decimal, other: Decimal): Decimal
	protected abstract fun updateValue(self: Double, other: Double): Double

	fun <T : Number> updateValue(self: T, other: T): T {
		return when (self) {
			is Decimal -> updateValue(self as Decimal, other as Decimal) as T
			is Double -> updateValue(self as Double, other as Double) as T
			is Int -> updateValue(self as Int, other as Int) as T
			else -> throw RuntimeException("Unknown number type: ${self::class.qualifiedName}")
		}
	}
}
